
(** Exception raised  when graphs g1 and g2 are not composible *)
exception Not_composible;;

(** verbose: flag that controls printing of intermediate computation *)
val verbose : bool ref 

(** variable identifier *)
type ident = string

(** type for a sub expression : a subexpression is a node number in the array 
    describing the numbered expression *)
type sub_expr = int

(** type for the endings of an arc of a size-change graph
  This is also the type of elements in the graph basis of a state.
  gb(e:p) = fv(e) U { * }
*)
type gb_element = Variable of int | Bullet

(** Label for the arcs of the size-change graph *)
type scg_arclabel = ArrowDown | ArrowEqual

(** type for a size-change graph arc *)
type scg_arc = gb_element * scg_arclabel * gb_element

(** type for control points *)
type ctlpt = int

(** Graph data structure. A graph is defined by:
  - a pair of control points defining the basis of the graph
  - a list of arcs
  - a list of subexpressions representing the program points that the graph passes through
  (this list is empty during the generation of sc graphs but it is used during the closure phase
  of the analysis: an element is added to this list if the corresponding graph is the 
  result of a graph composition)
*)
type scg = ctlpt * ctlpt * scg_arc list * ctlpt list

type scgs_mat = (scg_arc list * ctlpt list ) list array array
    
(** class for a program expression *)
class virtual ['tnode] tprogram : ident array * 'tnode array  ->
  (** 'tnode is the type of nodes in the abstract tree of the language 
    'tctrpoint is the type used for control points (used in judgement forms) *)
object
  
  (** [variables] is an array containing the different variable names appearing in the program expression *)
  val mutable variables : ident array 
    (** [expr] is an array containing the program subexpression nodes *)
  val mutable expr : 'tnode array 
    
  method virtual string_of_ctlpt : ctlpt -> string
    
  method print_scg_arcs : scg_arc list -> unit
  method print_scgraph : scg -> unit
  method print_scgraphs : scg list -> unit
    
  method eq_vertex :  gb_element -> gb_element -> bool
  method eq_arc :  scg_arc -> scg_arc -> bool
  method eq_arcs : scg_arc list -> scg_arc list -> bool    

  method scg_compose : scg_arc list -> scg_arc list -> scg_arc list

  method composition_closure :  scgs_mat -> unit
  method is_sct : scgs_mat  -> ( bool * (scg list) )

  method is_loop : scg -> bool
  method is_descending : scg -> bool
end 
  
  
module CtlPtSet : Set.S with type elt = ctlpt
			
class virtual ['tnode, 'jftype, 'jfgen] tgrgen_program :
  ident array * 'tnode array * CtlPtSet.t array ->
object
  inherit ['tnode] tprogram

  (* free variables  *)
  val mutable freevars : CtlPtSet.t array 
  method free_variables : ctlpt -> int list

  (* judgement forms *)
  val jf_array : ('jftype * 'jfgen) list array array
  val nb_jf : int ref

  method virtual string_of_jftype : 'jftype -> string
  method print_jf : ctlpt -> ctlpt -> 'jftype -> 'jfgen -> unit
  method create_jf : 'jftype -> ctlpt -> ctlpt -> 'jfgen -> unit
    
  (* generation component *)
  method virtual eq_gen : 'jfgen -> 'jfgen -> bool
  method virtual print_gen : 'jfgen -> unit

  (* miscellaneous *)
  method virtual is_closed : unit -> bool
  
  (* graph generation *)
  method gr_id_eq : ctlpt -> scg_arc list
  method gr_id_eq_subxx : int -> ctlpt -> scg_arc list
  method gr_id_des : ctlpt -> scg_arc list
  method gr_idv : ctlpt -> scg_arc list
  method gr_idv_subxx : int -> ctlpt -> scg_arc list

  method gr_var : int -> ctlpt -> scg_arc list
  method gr_var0 : int -> scg_arc list
  method gr_succ : scg_arc list -> scg_arc list
  method gr_pred : scg_arc list -> scg_arc list
  method gr_call : int -> scg_arc list -> scg_arc list -> scg_arc list
  method gr_call0 : int -> scg_arc list -> scg_arc list -> scg_arc list
  method gr_local : int -> scg_arc list -> ctlpt -> scg_arc list -> scg_arc list
  method gr_local0 : int -> scg_arc list -> ctlpt -> scg_arc list -> scg_arc list

end
			

